using System.Collections.Generic;
using System.Linq;
using System.Reflection;

namespace Upsilon.BaseTypes.UserData
{
    internal class UserDataType
    {
        public UserDataType(System.Type type)
        {
            Type = type;
            Variables = type.GetFields().ToDictionary(x => x.Name.ToLowerInvariant(), x => x);
            Properties = type.GetProperties().ToDictionary(x => x.Name.ToLowerInvariant(), x => x);
            Methods = new Dictionary<string, UserDataMethod>();
            foreach (var methodInfo in type.GetMethods())
            {
                var commonName = methodInfo.Name.ToLowerInvariant();
                if (Methods.TryGetValue(commonName, out var methodData))
                {
                    methodData.LoadMethodPart(methodInfo);
                }
                else
                {
                    Methods.Add(commonName, new UserDataMethod(methodInfo));
                }
            }
            OperatorHandler = new UserDataTypeOperators(type);
        }

        private System.Type Type { get; }
        private Dictionary<string, FieldInfo> Variables { get; }
        private Dictionary<string, PropertyInfo> Properties { get; }
        private Dictionary<string, UserDataMethod> Methods { get; }
        private UserDataTypeOperators OperatorHandler { get; }

        public (ScriptType Type, bool Failed, string Error) Get(object value, string member)
        {
            if (value.GetType() != Type)
                return (null, true, "Invalid Type");
            member = member.ToLowerInvariant();
            if (Variables.TryGetValue(member, out var info))
            {
                return (info.GetValue(value).ToScriptType(), false, null);
            }
            if (Properties.TryGetValue(member, out var property))
            {
                return (property.GetValue(value).ToScriptType(), false, null);
            }
            if (Methods.TryGetValue(member, out var method))
            {
                return (new ScriptMethodInfoFunction(method, value, false), false, null);
            }

            return (null, true, $"Can't find public member '{member}' on type '{Type}'.");
        }

        public (bool Failed, string Error) Set(object value, string member, ScriptType newValue)
        {
            member = member.ToLowerInvariant();
            if (value.GetType() != Type)
                return (true, "Invalid Type");
            if (Variables.TryGetValue(member, out var info))
            {
                info.SetValue(value, newValue.ToCSharpObject());
                return (false, null);
            }
            if (Properties.TryGetValue(member, out var property))
            {
                if (property.SetMethod == null || property.SetMethod.IsPrivate)
                {
                    return (true, $"Property '{member}' on type '{Type}' does not have a publicly available setter.");
                }
                property.SetValue(value, newValue.ToCSharpObject());
                return (false, null);
            }

            return (true, $"Cannot find member '{member}' on type '{Type}'");
        }

        public (ScriptType Type, bool Failed) BinaryOperator(object value, ScriptType par1, OperatorType op, ScriptType par2)
        {
            var method = OperatorHandler.GetBinaryOperator(op, par1.GetCSharpType(), par2.GetCSharpType());
            if (method == null)
            {
                return (new ScriptNull(), true);
            }

            return (method.Invoke(value, new[] {par1.ToCSharpObject(), par2.ToCSharpObject()}).ToScriptType(), false);
        }
        public (ScriptType Type, bool Failed) UnaryOperator(object value, ScriptType par1, OperatorType op)
        {
            var method = OperatorHandler.GetUnaryOperator(op, par1.GetCSharpType());
            if (method == null)
            {
                return (new ScriptNull(), true);
            }

            return (method.Invoke(value, new[] {par1.ToCSharpObject()}).ToScriptType(), false);
        }
    }
}