using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using Upsilon.BaseTypes;
using Upsilon.BaseTypes.Number;
using Upsilon.BaseTypes.ScriptTypeInterfaces;
using Upsilon.BaseTypes.UserData;
using Upsilon.Binder;
using Type = Upsilon.BaseTypes.Type;

namespace Upsilon.Evaluator
{
    internal class Evaluator : IDisposable
    {
        private Diagnostics _diagnostics;
        private ScriptType _lastValue;
        private ScriptType _returnValue;
        internal EvaluationScope Scope { get; private set; }
        private bool HasReturned { get; set; }
        private bool HasBroken { get; set; }

        internal Evaluator(Diagnostics diagnostics, Dictionary<string, ScriptType> variables)
        {
            _diagnostics = diagnostics;
            Scope       = new EvaluationScope(variables);
        }

        internal Evaluator(Diagnostics diagnostics, EvaluationScope parentScope)
        {
            _diagnostics = diagnostics;
            Scope        = new EvaluationScope(parentScope);
        }

        private Evaluator(){}

        internal static Evaluator CreateWithSetScope(Diagnostics diagnostics, EvaluationScope scope)
        {
            return new Evaluator {_diagnostics = diagnostics, Scope = scope};
        }

        public void Dispose()
        {
            Scope.Dispose();
            _lastValue = null;
            _returnValue = null;
        }

        public ScriptType Evaluate(BoundScript e)
        {
            EvaluateNode(e.Statement);
            if (_returnValue == null)
                return _lastValue;
            return _returnValue;
        }

        public ScriptType Evaluate(BoundScript e, string functionName, ImmutableArray<ScriptType> parameters)
        {
            EvaluateNode(e.Statement);
            if (!Scope.TryGet(functionName, out var statement) || statement.Type != Type.Function)
            {
                throw new ArgumentException(($"Function '{functionName}' could not be found"));
            }
            var function = (ScriptInternalFunction) statement;
            var innerEvaluator = new Evaluator(_diagnostics, Scope);
            for (var index = 0; index < parameters.Length; index++)
            {
                var parameter = parameters[index];
                var parameterName = function.Parameters[index];
                innerEvaluator.Scope.CreateLocal(parameterName.VariableSymbol, parameter);
            }

            var result = innerEvaluator.EvaluateNode(function.Block);
            return result;
        }

        internal ScriptType EvaluateNode(BoundNode b)
        {
            switch (b.Kind)
            {
                case BoundKind.BoundScript:
                    EvaluateStatement(((BoundScript)b).Statement);
                    break;
                case BoundKind.BoundLiteralExpression:
                case BoundKind.BoundBinaryExpression:
                case BoundKind.BoundUnaryExpression:
                case BoundKind.VariableExpression:
                case BoundKind.BoundFunctionCallExpression:
                case BoundKind.BoundFunctionExpression:
                case BoundKind.BoundTableExpression:
                case BoundKind.BoundIndexExpression:
                case BoundKind.BoundFullstopIndexExpression:
                    _lastValue = EvaluateExpression((BoundExpression) b);
                    break;
                case BoundKind.BoundAssignmentStatement:
                case BoundKind.BoundExpressionStatement:
                case BoundKind.BoundBlockStatement:
                case BoundKind.BoundIfStatement:
                case BoundKind.BoundElseStatement:
                case BoundKind.BoundFunctionAssignmentStatement:
                case BoundKind.BoundPromise:
                case BoundKind.BoundReturnStatement:
                case BoundKind.BoundTableAssigmentStatement:
                case BoundKind.BoundMultiAssignmentStatement:
                case BoundKind.BoundNumericForStatement:
                case BoundKind.BoundGenericForStatement:
                case BoundKind.BoundBreakStatement:
                    EvaluateStatement((BoundStatement) b);
                    break;
                default:
                    throw new ArgumentOutOfRangeException();
            }
            return _returnValue;
        }

        private void EvaluateStatement(BoundStatement e)
        {
            if (HasReturned)
                return;
            switch (e.Kind)
            {
                case BoundKind.BoundAssignmentStatement:
                    EvaluateAssignmentStatement((BoundVariableAssignment) e);
                    break;
                case BoundKind.BoundBlockStatement:
                    EvaluateBoundBlockStatement((BoundBlockStatement) e);
                    break;
                case BoundKind.BoundIfStatement:
                    EvaluateBoundIfStatement((BoundIfStatement) e);
                    break;
                case BoundKind.BoundReturnStatement:
                    EvaluateReturnStatement((BoundReturnStatement) e);
                    break;
                case BoundKind.BoundFunctionAssignmentStatement:
                    EvaluateBoundFunctionAssigmentStatement((BoundFunctionAssignmentStatement) e);
                    break;
                case BoundKind.BoundTableAssigmentStatement:
                    EvaluateTableAssignmentStatement((BoundTableAssigmentStatement) e);
                    break;
                case BoundKind.BoundMultiAssignmentStatement:
                    EvaluateMultiAssignmentStatement((BoundMultiAssignmentStatement) e);
                    break;
                case BoundKind.BoundNumericForStatement:
                    EvaluateNumericForStatement((BoundNumericForStatement) e);
                    break;
                case BoundKind.BoundBreakStatement:
                    HasBroken = true;
                    return;
                case BoundKind.BoundGenericForStatement:
                    EvaluateGenericForStatement((BoundGenericForStatement) e);
                    break;
                default:
                    EvaluateExpressionStatement((BoundExpressionStatement) e);
                    break;
            }
        }


        private void EvaluateExpressionStatement(BoundExpressionStatement e)
        {
            _lastValue = EvaluateExpression(e.Expression);
        }

        internal ScriptType EvaluateExpression(BoundExpression e)
        {
            switch (e.Kind)
            {
                case BoundKind.BoundLiteralExpression:
                    return ((BoundLiteralExpression) e).Value;
                case BoundKind.BoundBinaryExpression:
                    return EvaluateBinaryExpression((BoundBinaryExpression) e);
                case BoundKind.BoundUnaryExpression:
                    return EvaluateUnaryExpression((BoundUnaryExpression) e);
                case BoundKind.VariableExpression:
                    return EvaluateVariableExpression((BoundVariableExpression) e);
                case BoundKind.BoundFunctionCallExpression:
                    return EvaluateBoundFunctionCallExpression((BoundFunctionCallExpression) e);
                case BoundKind.BoundTableExpression:
                    return EvaluateTableExpression((BoundTableExpression) e);
                case BoundKind.BoundIndexExpression:
                    return EvaluateIndexExpression((BoundIndexExpression) e);
                case BoundKind.BoundFunctionExpression:
                    return EvaluateBoundFunctionStatement((BoundFunctionExpression) e);
                case BoundKind.BoundPromise:
                    return EvaluateUnboundFunctionStatement((UnboundFunctionExpression) e);
                case BoundKind.BoundFullstopIndexExpression:
                    return EvaluateFullStopIndexExpression((BoundFullStopIndexExpression) e);
                default:
                    throw new NotImplementedException();
            }
        }

        private ScriptType EvaluateUnaryExpression(BoundUnaryExpression e)
        {
            var operand  = EvaluateExpression(e.InExpression);
            switch (e.Operator.Kind)
            {
                case BoundUnaryOperator.OperatorKind.Identity:
                    return operand;
                case BoundUnaryOperator.OperatorKind.Negation:
                    if (operand.Type == Type.Number)
                        return -((ScriptNumber)operand);
                    else if (operand.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) operand;
                        var (type, failed) = ud.UnaryOperator(operand, OperatorType.UnaryNegation);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundUnaryOperator.OperatorKind.LogicalNegation:
                    if (operand.Type == Type.Boolean)
                        return !(ScriptBoolean) operand;
                    else if (operand.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) operand;
                        var (type, failed) = ud.UnaryOperator(operand, OperatorType.LogicalNot);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundUnaryOperator.OperatorKind.Length:
                    if (operand is ILengthType length)
                    {
                        return length.Length();
                    }
                    goto default;
                default:
                    throw new Exception("Invalid Unary Operator: " + e.Operator.Kind);
            }
        }

        private ScriptType EvaluateBinaryExpression(BoundBinaryExpression e)
        {
            var left  = EvaluateExpression(e.LeftExpression);
            var right = EvaluateExpression(e.RightExpression);
            switch (e.Operator.Kind)
            {
                case BoundBinaryOperator.OperatorKind.Addition:
                    if (left.Type == Type.Number)
                    {
                        if (right.Type == Type.Number)
                        {
                            return ((ScriptNumber)left) + ((ScriptNumber)right);
                        }
                        goto default;
                    }
                    else if (left.Type == Type.String)
                    {
                        return ((ScriptString) left) + right;
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Addition, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Subtraction:
                    if (left.Type == Type.Number)
                    {
                        return ((ScriptNumber)left) - ((ScriptNumber)right);
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Subtraction, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Multiplication:
                    if (left.Type == Type.Number)
                    {
                        return ((ScriptNumber)left) * ((ScriptNumber)right);
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Multiplication, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Division:
                    if (left.Type == Type.Number)
                    {
                        return ((ScriptNumber)left) / ((ScriptNumber)right);
                    }
                    else if (left.Type == Type.UserData)
                    {
                        var ud = (GenericUserData) left;
                        var (type, failed) = ud.BinaryOperator(left, OperatorType.Division, right);
                        if (failed) goto default;
                        return type;
                    }
                    goto default;
                case BoundBinaryOperator.OperatorKind.Equality:
                    return new ScriptBoolean(Equals(left, right));
                case BoundBinaryOperator.OperatorKind.Inequality:
                    return new ScriptBoolean(!Equals(left, right));
                default:
                    throw new Exception("Invalid Binary Operator: " + e.Operator);
            }
        }

        private void EvaluateAssignmentStatement(BoundVariableAssignment e)
        {
            var val = EvaluateExpression(e.BoundExpression);
            if (e.IsLocalDefinition)
            {
                Scope.CreateLocal(e.Variable.VariableSymbol, val);
            }
            else
            {
                Scope.AssignToNearest(e.Variable.VariableSymbol, val);
            }
            _lastValue = val;
        }

        private void EvaluateMultiAssignmentStatement(BoundMultiAssignmentStatement e)
        {
            var val = EvaluateExpression(e.Assignment);
            if (val is IIndexable table)
            {
                for (var i = 0; i < e.Variables.Length; i++)
                {
                    var variableSymbol = e.Variables[i];
                    if (variableSymbol == null)
                        continue;
                    if (variableSymbol.Name == "_")
                        continue;
                    var value = table.Get(_diagnostics, e.Span, new ScriptNumberLong(i + 1), Scope);
                    if (variableSymbol.Local)
                        Scope.CreateLocal(variableSymbol, value);
                    else
                        Scope.AssignToNearest(variableSymbol, value);
                }
            }
            else
            {
                _diagnostics.LogError($"Can't assign type '{val.Type}' to multiple variables.", e.Span);
            }
            _lastValue = val;
        }

        private ScriptType EvaluateVariableExpression(BoundVariableExpression e)
        {
            if (Scope.TryGet(e.Variable.VariableSymbol, out var val))
            {
                return val;
            }
            throw new Exception($"Cannot find variable: '{e.Variable.VariableSymbol.Name}'");
        }

        private void EvaluateBoundBlockStatement(BoundBlockStatement boundBlockStatement)
        {
            foreach (var boundStatement in boundBlockStatement.Statements)
            {
                if (HasReturned)
                    return;
                if (HasBroken)
                    return;
                EvaluateStatement(boundStatement);
            }
        }

        private void EvaluateBoundIfStatement(BoundIfStatement boundBlockStatement)
        {
            var condition = EvaluateExpression(boundBlockStatement.Condition.Expression);
            if ((ScriptBoolean) condition)
            {
                EvaluateStatement(boundBlockStatement.Block);
            }
            else if (boundBlockStatement.NextElseIf != null)
            {
                EvaluateStatement(boundBlockStatement.NextElseIf);
            }
            else if (boundBlockStatement.ElseStatement != null)
            {
                EvaluateStatement(boundBlockStatement.ElseStatement.Block);
            }
        }

        private void EvaluateBoundFunctionAssigmentStatement(BoundFunctionAssignmentStatement e)
        {
            var func = EvaluateBoundFunctionStatement(e.Func);
            if (e.Variable.Local)
                Scope.CreateLocal(e.Variable, func);
            else
                Scope.AssignToNearest(e.Variable, func);
        }

        private ScriptType EvaluateBoundFunctionStatement(BoundFunctionExpression boundFunctionExpression)
        {
            var func = new ScriptInternalFunction(boundFunctionExpression.Parameters, boundFunctionExpression.Block, Scope);
            return func;
        }

        private ScriptType EvaluateUnboundFunctionStatement(UnboundFunctionExpression unboundFunctionExpression)
        {
            var func = new ScriptInternalFunction(unboundFunctionExpression.Parameters, unboundFunctionExpression.Block, Scope);
            return func;
        }

        private ScriptType EvaluateBoundFunctionCallExpression(BoundFunctionCallExpression boundFunctionCallExpression)
        {
            var variable = EvaluateExpression(boundFunctionCallExpression.Identifier);
            if (!(variable is ScriptFunction function))
            {
                throw new Exception($"Variable is not a function.");
            }
            var ls = new List<ScriptType>();
            foreach (var t in boundFunctionCallExpression.Parameters)
            {
                var evaluate = EvaluateExpression(t);
                ls.Add(evaluate);
            }

            var val = function.Run(_diagnostics, ls.ToArray());
            return val;
        }

        private void EvaluateReturnStatement(BoundReturnStatement b)
        {
            _returnValue = EvaluateExpression(b.Expression);
            _lastValue = _returnValue;
            HasReturned = true;
        }

        private ScriptType EvaluateTableExpression(BoundTableExpression e)
        {
            var tableScope = EvaluationScope.CreateWithGetOnlyParent(Scope);
            var innerEvaluator = new Evaluator(_diagnostics, tableScope);
            var currentPos = 1;
            foreach (var boundStatement in e.Statements)
            {
                switch (boundStatement.Kind)
                {
                    case BoundKind.BoundAssignmentStatement:
                    case BoundKind.BoundFunctionExpression:
                    case BoundKind.BoundFunctionAssignmentStatement:
                        innerEvaluator.EvaluateNode(boundStatement);
                        break;
                    default:
                    {
                        innerEvaluator.EvaluateStatement(boundStatement);
                        if (innerEvaluator._lastValue != null)
                            tableScope.CreateLocal(new VariableSymbol(currentPos.ToString(), innerEvaluator._lastValue.Type, false),
                                innerEvaluator._lastValue);
                        innerEvaluator._lastValue = null;
                        break;
                    }
                }

                currentPos++;
            }
            return new ScriptTable(tableScope);
        }

        private ScriptType EvaluateIndexExpression(BoundIndexExpression e)
        {
            var variable = EvaluateExpression(e.Identifier);
            if (!(variable is IIndexable indexable))
            {
                throw new Exception("Variable is not indexable.");
            }

            var scope = Scope;
            if (variable is IScopeOwner scopeOwner)
            {
                scope = scopeOwner.EvaluationScope;
            }
            var indexer   = EvaluateExpression(e.Index);
            return indexable.Get(_diagnostics, e.Span, indexer, scope);
        }
        private ScriptType EvaluateFullStopIndexExpression(BoundFullStopIndexExpression e)
        {
            var variable = EvaluateExpression(e.Expression);
            if (!(variable is IIndexable indexable))
            {
                throw new Exception("Variable is not indexable.");
            }

            var scope = Scope;
            if (variable is IScopeOwner scopeOwner)
            {
                scope = scopeOwner.EvaluationScope;
            }

            return indexable.Get(_diagnostics, e.Span, e.Index.ToScriptType(), scope);
        }

        private void EvaluateTableAssignmentStatement(BoundTableAssigmentStatement e)
        {
            ScriptType table;
            ScriptType index;
            if (e.TableIndexExpression.Kind == BoundKind.BoundIndexExpression)
            {
                table = EvaluateExpression(((BoundIndexExpression)e.TableIndexExpression).Identifier);
                index = EvaluateExpression(((BoundIndexExpression)e.TableIndexExpression).Index);
            }
            else
            {
                table = EvaluateExpression(((BoundFullStopIndexExpression)e.TableIndexExpression).Expression);
                index = ((BoundFullStopIndexExpression) e.TableIndexExpression).Index.ToScriptType();
            }

            var value = EvaluateExpression(e.Value);
            if (!(table is IIndexable indexable))
            {
                throw new Exception("Cant assign to that");
            }
            indexable.Set(_diagnostics, e.Span, index.ToString().ToScriptType(), value);
        }

        private void EvaluateNumericForStatement(BoundNumericForStatement e)
        {
            var innerEvaluator = new Evaluator(_diagnostics, Scope);
            var startVal = (ScriptNumberLong)innerEvaluator.EvaluateExpression(e.BoundStart);
            innerEvaluator.Scope.CreateLocal(e.Variable, startVal);
            var stopVal = (ScriptNumberLong)innerEvaluator.EvaluateExpression(e.BoundStop);
            long step = 1;
            if (e.BoundStep != null)
            {
                var stepVal = (ScriptNumberLong)innerEvaluator.EvaluateExpression(e.BoundStep);
                step = stepVal.Value;
            }

            if (step > 0)
            {
                for (; startVal.Value <= stopVal.Value; startVal.Value = startVal.Value + step)
                {
                    innerEvaluator.EvaluateBoundBlockStatement(e.Block);
                    if (innerEvaluator.HasBroken)
                        break;
                }
            }
            else if (step < 0)
            {
                for (; startVal.Value >= stopVal.Value; startVal.Value = startVal.Value + step)
                {
                    innerEvaluator.EvaluateBoundBlockStatement(e.Block);
                    if (innerEvaluator.HasBroken)
                        break;
                }
            }
        }

        private void EvaluateGenericForStatement(BoundGenericForStatement e)
        {
            var innerEvaluator = new Evaluator(_diagnostics, Scope);
            var enumeratorObject = EvaluateExpression(e.BoundEnumerableExpression);
            if (!(enumeratorObject is IIterable iterable))
            {
                throw new Exception($"Can't iterate over object with type '{enumeratorObject.Type}'");
            }

            using (var enumerator = iterable.GetEnumerator())
            {
                while (enumerator.MoveNext())
                {
                    var current = enumerator.Current;
                    if (current == null || current.Type != Type.Table)
                    {
                        throw new Exception($"Can't assign result value with type '{current.Type}' to multiple values");
                    }

                    var table = (SimpleScriptTable)current;
                    if (e.Variables[0].Name != "_")
                        innerEvaluator.Scope.CreateLocal(e.Variables[0], table[0].ToScriptType());
                    if (e.Variables[1].Name != "_")
                        innerEvaluator.Scope.CreateLocal(e.Variables[1], table[1]);
                    innerEvaluator.EvaluateBoundBlockStatement((BoundBlockStatement) e.Block);
                    if (innerEvaluator.HasBroken)
                        break;
                }
            }
        }
    }
}