using System;
using Upsilon.BaseTypes.UserData;
using Upsilon.Evaluator;
using Xunit;

// ReSharper disable UnusedMember.Local

namespace UpsilonTests.GeneralTests
{
    public class UserDataTests : TestClass, IClassFixture<UserDataTests.UserDataTestsFixture>
    {
        public class UserDataTestsFixture : IDisposable
        {
            public UserDataTestsFixture()
            {
                UserDataTypeHandler.LoadType<UserDataHelper>();
            }

            public void Dispose()
            {
            }
        }

        #pragma warning disable 414, 649
        private class UserDataHelper
        {
            public string FieldString = "TestField";
            public string FieldStringSet;
            private string _privateTestField = "hidden";
            public bool GetOnly { get; } = false;
            public bool PrivateSet { get; private set; } = false;

            public bool TestMethodHasRun { get; private set; }
            public void TestMethod()
            {
                TestMethodHasRun = true;
            }

            public long Add(long a, long b)
            {
                return a + b;
            }

        }
        #pragma warning restore 414, 649


        [Fact]
        public void AccessFieldsGet()
        {
            var obj = new UserDataHelper();
            const string input = @"
function test(o)
    return o.fieldString
end
";
            var script = new Script(input, BoundScope, StaticScope);
            Assert.Empty(script.Diagnostics.Messages);
            var result = script.EvaluateFunction<string>("test", new[] {obj});
            Assert.Empty(script.Diagnostics.Messages);
            Assert.Equal("TestField", result);
        }

        [Fact]
        public void AccessFieldsSet()
        {
            var obj = new UserDataHelper();
            const string input = @"
function test(o)
    o.FieldStringSet = ""Test""
end
";
            var script = new Script(input, BoundScope, StaticScope);
            Assert.Empty(script.Diagnostics.Messages);
            script.EvaluateFunction("test", new[] {obj});
            Assert.Empty(script.Diagnostics.Messages);
            Assert.Equal("Test", obj.FieldStringSet);
        }

        [Fact]
        public void RunVoidMethod()
        {
            var obj = new UserDataHelper();
            const string input = @"
function test(o)
    o.testMethod()
end
";
            var script = new Script(input, BoundScope, StaticScope);
            Assert.Empty(script.Diagnostics.Messages);
            script.EvaluateFunction("test", new[] {obj});
            Assert.Empty(script.Diagnostics.Messages);
            Assert.True(obj.TestMethodHasRun);
        }

        [Fact]
        public void ReturnMethodWithParameters()
        {
            var obj = new UserDataHelper();
            const string input = @"
function test(o)
    return o.add(100, 20)
end
";
            var script = new Script(input, BoundScope, StaticScope);
            Assert.Empty(script.Diagnostics.Messages);
            var result = script.EvaluateFunction<long>("test", new[] {obj});
            Assert.Empty(script.Diagnostics.Messages);
            Assert.Equal(120, result);
        }

        [Fact]
        public void CantAccessHiddenFields()
        {
            var obj = new UserDataHelper();
            const string input = @"
function test(o)
    return o._privateTestField
end
";
            var script = new Script(input, BoundScope, StaticScope);
            Assert.Empty(script.Diagnostics.Messages);
            script.EvaluateFunction("test", new[] {obj});
            Assert.Single(script.Diagnostics.Messages);
        }

        [Fact]
        public void CantSetToFieldsWithNoSetter()
        {
            var obj = new UserDataHelper();
            const string input = @"
function test(o)
    o.GetOnly = true
end
";
            var script = new Script(input, BoundScope, StaticScope);
            Assert.Empty(script.Diagnostics.Messages);
            script.EvaluateFunction("test", new[] {obj});
            Assert.Single(script.Diagnostics.Messages);
            Assert.False(obj.GetOnly);
        }

        [Fact]
        public void CantSetToFieldsWithPrivateSetter()
        {
            var obj = new UserDataHelper();
            const string input = @"
function test(o)
    o.PrivateSet = true
end
";
            var script = new Script(input, BoundScope, StaticScope);
            Assert.Empty(script.Diagnostics.Messages);
            script.EvaluateFunction("test", new[] {obj});
            Assert.Single(script.Diagnostics.Messages);
            Assert.False(obj.PrivateSet);
        }

        public UserDataTests(StaticScriptFixture fix) : base(fix)
        {
        }
    }
}